//--------------------------------------------------------------------------
// Copyright (c) 1999,2000 Southeastern Universities Research Association,
//                         Continuous Electron Beam Accelerator Facility
//
// This software was developed under a United States Government license
// described in the NOTICE file included as part of this distribution.
//--------------------------------------------------------------------------
//
//  Description:
//      This is a class that extends JFrame. This class is a utility class
//      for CTable. It has file management capabilities. Most of all it is  
//      to be able to use the table inside of its container.
//
//  Author: Masa Ebisawa
//
//  Based on work by Chip Watson
//                                       
//--------------------------------------------------------------------------
// Revision History:
//  05/??/00    Release 0.0  Masa Ebisawa 
//  07/06/00    Release 0.1  Reorganize the code (yc) 
//
//--------------------------------------------------------------------------

package sdm.apps;

import java.awt.*;
import java.awt.event.*;
import java.io.*;
import javax.swing.*;
import javax.swing.table.*;
import javax.swing.event.*;
import javax.swing.JFileChooser.*;
import java.util.*;
import java.beans.*;
import java.awt.print.*;

import sdm.*;
import cdev.*;
import cdev.data.*;


/**
 * This class is a utility class for CTable. It is so that you can use 
 * the table and be able to store the configuration and be able to view
 * multiple monitors at the same time with one utility. This class also 
 * implements PropertyChangeListener that can detect if a property has 
 * changed in another class.  
 */
public class CTableUtility extends JFrame implements Printable, 
			    ActionListener, PropertyChangeListener { 

    private JMenuItem newTable = new JMenuItem("New");
    private JMenuItem openTable = new JMenuItem("Open");
    private JMenuItem saveTable = new JMenuItem("Save");
    private JMenuItem saveAsTable = new JMenuItem("SaveAs");
    private JMenuItem close = new JMenuItem("Close");
    private JMenuItem printMenu = new JMenuItem("Print");
    private JMenuItem suspend = new JMenuItem("Pause");
    private JMenuItem resume = new JMenuItem("Start"); 
    protected boolean change;
    protected JFileChooser fch;
    protected CTable table;
    protected Container contentPane;
    protected JScrollPane scroll;
    protected ControlSystem sys;
    protected String name;
    protected newInputFrame frame;	
    protected int lastSizeOfColumns, sizeChanged;
    protected String openFile;

    
    /**
     *Creates a frame with file management capabilities
     *
     *@param fileName     The name of the file to be read in   
     */
    public CTableUtility (String fileName) {
	create("localhost", 9200, fileName);
    }

    /**
     *Creates a frame with file management capabilities
     *
     *@param host         The name of the host to be connected to.        
     *@param port         The port number to connect to.         
     */
    public CTableUtility (String host, String port) {
	int portNumber = Integer.parseInt(port);
	create(host, portNumber, null);
    }

    /**
     *Creates a frame with file management capabilities
     *
     *@param host         The name of the host to be connected to.        
     *@param port         The port number to connect to.         
     *@param fileName     The name of the file to be read in.   
     */
    public CTableUtility(String host, String port, String fileName) {
	int portNumber = Integer.parseInt(port);
	create(host, portNumber, fileName);
    }

    private void create(String host, int port, String fileName){
	setTitle ("CTable");
	setSize (400, 300);	
	addWindowListener (new WindowAdapter() {
	    public void windowClosing (WindowEvent event) {
		closeTable ();
	    }
	});
	try { 
	    sys = ControlSystem.getDefault();
	    sys.connect(host, port);	    
	} catch(Exception e){
	    e.printStackTrace();
	}
	contentPane = getContentPane();
	JMenuBar menuBar = new JMenuBar();
	setJMenuBar(menuBar);
	JMenu options = new JMenu("Files");
	JMenu s = new JMenu("System");
	menuBar.add(options);
	menuBar.add(s);

	options.add(newTable);
	options.add(openTable);
	options.add(saveTable);
	options.add(saveAsTable);
	options.addSeparator();
	options.add(printMenu);
	options.addSeparator();
	options.add(close);
	s.add(suspend);
	s.add(resume);
	change = false;
	newTable.addActionListener(this);		
	saveTable.addActionListener(this);
	saveAsTable.addActionListener(this);
	openTable.addActionListener(this);
	printMenu.addActionListener(this);
	close.addActionListener(this);
	suspend.addActionListener(this);
	resume.addActionListener(this);
	
	sizeChanged = 0;
	lastSizeOfColumns = -1;
	if(fileName != null) {
	    openFile = fileName;
	    openMethod();
	}
    }
    
    /**
     * This is a function for preopertyChangelistener
     *
     * @param evt catches the PropertyChangeEvent     
     */
    public void propertyChange(PropertyChangeEvent evt){
	table.setPreferredScrollableViewportSize(table.getPreferredSize()); 
	validate();
	pack();
    }
    
    
    public void actionPerformed(ActionEvent event) {
	if (event.getSource() == newTable )        newTable ();
	else if (event.getSource() == saveTable)   saveTable ();
	else if (event.getSource() == close ) 	   closeTable (); 
	else if (event.getSource() == saveAsTable) {
	    if (table != null) {
		if (fch == null)  fch = new JFileChooser(); 
		if (fch.showSaveDialog(this) == JFileChooser.APPROVE_OPTION)
		    writeToFile(); 
	    } 
	    table.getTableHeader().repaint();
	}
	else if (event.getSource() == openTable) {
	    if (table != null && table.getTableState()) { 
		String msg = "Do you want to create a new table?";
		int n = optionPanel(msg, "Confirming", 2);
		if (n == JOptionPane.YES_OPTION) { 
		    openSave();
		    openMethod();
		    validate();
		}
	    }
	    else {
		openMethod();
		validate();
	    }
	}
	else if (event.getSource() == printMenu ) {
	    PrinterJob pj = PrinterJob.getPrinterJob();
	    pj.setPrintable (this);
	    if (pj.printDialog()) {
		try { pj.print(); }
		catch (Exception PrintException) { }
	    }
	}
	else if (event.getSource() == suspend) {
	    try { sys.suspendIO();  }
	    catch(Exception e) { e.printStackTrace(); }
	}	
	else if (event.getSource() == resume) {
	    try { sys.resumeIO();   }
	    catch(Exception e) {  e.printStackTrace(); }
	}	
    }
	
    private void newTable() {
	if (table == null) { 
	    frame = new newInputFrame();
	    try {
		frame.show();
		frame.devices.requestFocus();
	    } catch (Exception e) { e.printStackTrace();  }	    
	}
	else {
	    String msg = "Do you want to create a new table?";
	    int n = optionPanel(msg, "Confirming", 3);
	    if (n == JOptionPane.YES_OPTION) {
		openSave();
		frame = new newInputFrame();
		frame.show();
		frame.devices.requestFocus();
	    }
	}
	name = "Default";
	validate();
    }

    /*
     * id == 1: showMessageDialog (QUESTION_MESSAGE)
     * id == 2: showMessageDialog (ERROR_MESSAGE)
     * id == 3: showConfirmDialog
     */
    private int optionPanel(String msg, String title, int id) {
	int n = -1;
	if (id == 1)
	    JOptionPane.showMessageDialog(frame, msg, title,
					  JOptionPane.QUESTION_MESSAGE);
	else if (id == 2)
	    JOptionPane.showMessageDialog(frame, msg, title,
					  JOptionPane.ERROR_MESSAGE);
	else if (id == 3) 
	    n=JOptionPane.showConfirmDialog(frame, msg, title,
					    JOptionPane.YES_NO_OPTION,
					    JOptionPane.QUESTION_MESSAGE);
	return n;
    }

    /*
     * Write the table config info as objects into file
     */
    private void writeToFile() {
	try {
	    File file = fch.getSelectedFile();	   
	    ObjectOutputStream out = new 
		   ObjectOutputStream(new FileOutputStream(file));
	    int col = table.getPropertyNames().length - 1;
	    int row = table.getDeviceNames().length;
	    out.writeObject(new Dimension(col, row));
	    out.writeObject(table.getDeviceNames());
	    out.writeObject(table.getPropertyNames());
	    out.writeObject(table.getColumnWidths());
	    out.close();
	} catch(Exception e) {
	    e.printStackTrace();
	    System.exit(1);
	}    
    }

    private void saveTable () {
	if (table != null && table.getTableState()) {
	    if (fch == null) {
		fch = new JFileChooser();
		if (fch.showSaveDialog (this) ==
		    JFileChooser.APPROVE_OPTION) 
		writeToFile();
	    } 
	    else  { writeToFile(); }
	    change = false;
	}
    }

    /*
     * This function closes the application
     */
    private void closeTable () {
	if (table != null && table.getTableState()) {
	    String msg = "Do you want to save before leaving ?";
	    int n = optionPanel(msg, "Confirming", 3); 
	    if (n == JOptionPane.YES_OPTION) {
		if (fch==null)   fch = new JFileChooser();
		if (fch.showSaveDialog (this) ==
		    JFileChooser.APPROVE_OPTION) 
		writeToFile();
	    }
	}
	dispose();
	System.exit(0);
    }
   
    /*
     * This function saves whenever something is closed to makeway
     * for the new table.
     */
    private void openSave() {
	if (fch == null)  fch = new JFileChooser();
	if (fch.showSaveDialog (this) == JFileChooser.APPROVE_OPTION) 
	    writeToFile();
	scroll.removeAll();
	contentPane.removeAll();
	requestFocus();
	table = null;
    }

    /**
     *<!-- print(Graphic, PageFormat, pageIndex) -->
     *
     * This method is used to print the table.
     *
     * @param g   gets the graphics for the table
     * @param pageFormat    gets the page format for the printed page
     * @param pageIndex     gets the index of the page to be printed  
     */
    public int print(Graphics g, PageFormat pageFormat, int pageIndex) {
	Graphics2D g2 = (Graphics2D) g;
	g2.setColor(Color.black);
	int fh = g2.getFontMetrics().getHeight();
	int fd = g2.getFontMetrics().getDescent();

	//--- leave room for page number ------------
	double pageH = pageFormat.getImageableHeight() - fh;
	double pageW = pageFormat.getImageableWidth();
	double tableW = (double)table.getColumnModel().getTotalColumnWidth();
	double scale = 1; 
	if (tableW >= pageW)  scale = pageW / tableW;
	
	double headerHeightOnPage = table.getTableHeader().getHeight()*scale;
	double tableWidthOnPage = tableW * scale;
	
	double oneRowHeight=(table.getRowHeight()+
			     table.getRowMargin())*scale;
	int numRowsOnAPage=
	    (int)((pageH - headerHeightOnPage)/oneRowHeight);
	double pageHeightForTable=oneRowHeight * numRowsOnAPage;
	int totalNumPages = (int)Math.ceil(((double)
			   table.getRowCount())/numRowsOnAPage);
	if (pageIndex >= totalNumPages)  return NO_SUCH_PAGE;
	
	g2.translate(pageFormat.getImageableX(), 
		     pageFormat.getImageableY());
	//--- bottom center ----------------------------
	g2.drawString("Page: "+(pageIndex+1),
		      (int)pageW/2-35, (int)(pageH+fh-fd));
	
	g2.translate(0f, headerHeightOnPage);
	g2.translate(0f, -pageIndex*pageHeightForTable);
	
	//--- If this piece of the table is smaller than the size 
	// available, clip to the appropriate bounds. -------------
	if (pageIndex + 1 == totalNumPages) {
	    int lastRowPrinted = numRowsOnAPage * pageIndex;
	    int numRowsLeft = table.getRowCount() - lastRowPrinted;
	    g2.setClip(0, (int)(pageHeightForTable * pageIndex),
		          (int) Math.ceil(tableWidthOnPage),
		          (int) Math.ceil(oneRowHeight * numRowsLeft));
	}
	//--- else clip to the entire area available -----------
	else {    
	    g2.setClip(0, (int)(pageHeightForTable*pageIndex), 
		          (int) Math.ceil(tableWidthOnPage),
		          (int) Math.ceil(pageHeightForTable));        
	}
	
	g2.scale(scale,scale);
	table.paint(g2);
	g2.scale(1/scale, 1/scale);
	g2.translate(0f, pageIndex*pageHeightForTable);
	g2.translate(0f, -headerHeightOnPage);
	g2.setClip(0, 0, (int)Math.ceil(tableWidthOnPage), 
		         (int)Math.ceil(headerHeightOnPage));
	g2.scale(scale, scale);
	table.getTableHeader().paint(g2);
	//--- paint header at top --------------------
	return Printable.PAGE_EXISTS;
    }
    
    /**
     * due to a probable error in swing we needed to add this function
     * to repaint the table header of CTable when things are updated.
     */
     public void repaint (long l, int x, int y, int w, int h) {
    	super.repaint(l, x, y, w, h);
    	if (table!=null) {
    	    table.getTableHeader().repaint();
    	}
    }
    
    
    private void readFromFile (ObjectInputStream in) {
	try {
	    Dimension d = (Dimension)in.readObject();
	    String[] device = new String[d.height]; 
	    String[] property = new String[d.width];
	    int[] columnWidth = new int[d.width];
	    device = (String[])in.readObject(); 
	    property = (String[])in.readObject();
	    columnWidth = (int[])in.readObject();
	    in.close();
	    table = new CTable(device, property, columnWidth);
	    table.addPropertyChangeListener(this);
	    scroll = new JScrollPane(table); 
	    scroll.setViewportView(table);
	    contentPane.add(scroll, BorderLayout.CENTER);
	    validate();
	    table.setPreferredScrollableViewportSize(table.getPreferredSize()); 
	    scroll.updateUI();
	    validate();
	    pack();
	} catch(Exception e) { e.printStackTrace(); }
    }
    
    /*
     * this method opens a table from the indicated file
     */
    private void openMethod() {
	try {
	    if (fch == null)  fch = new JFileChooser();
	    if (openFile == null) {
		int returnval = fch.showOpenDialog (this);
		if (returnval == JFileChooser.APPROVE_OPTION) {
		    File input = fch.getSelectedFile();
		    ObjectInputStream in = new
			ObjectInputStream (new FileInputStream(input));
		    readFromFile (in);
		}
	    }
	    else { 
		ObjectInputStream in = new
		    ObjectInputStream (new FileInputStream(openFile));
		readFromFile (in);
		openFile = null; 
	    }
	} catch (Exception e) {
	    if (e instanceof StreamCorruptedException) {
		String msg = "Invalid File \n"+fch.getSelectedFile();
		optionPanel(msg, "Invalid File", 1);
	    }
	    else if (e instanceof FileNotFoundException) {
		String msg = "File "+fch.getSelectedFile()+"\ndoes not exist";
		optionPanel(msg, "No such File", 1);
	    }
	    else { e.printStackTrace();	}
	    if (this != null)      openMethod();  //????
	    if (this.isVisible())  openMethod();
	    else                   openFile = null;
	}
    }
    
    private class newInputFrame extends JFrame implements ActionListener {

	protected JButton cancel = new JButton("CANCEL");
        protected JButton ok = new JButton("ENTER");
	protected JTextField devices = new JTextField(80);
	protected JTextField properties = new JTextField(80);
	protected String[] dName;
	protected String[] pName; 

	public newInputFrame() {
	    setTitle("Create New Table");
	    setSize(420, 150);

	    //--- text Panel ----------------------------
	    JPanel textP = new JPanel();
	    textP.setLayout(new GridLayout(4, 1));
	    devices.addActionListener(this);
	    String msg1 = "List of Device (Seperated by space or a comma)";
	    textP.add(new JLabel(msg1));
	    textP.add(devices);
	    properties.addActionListener(this);
	    String msg2 = "List of Properties (Seperated by space or a comma)";
	    textP.add(new JLabel(msg2));
	    textP.add(properties);

	    //--- button Panel --------------------------
	    JPanel buttonP = new JPanel();
	    ok.addActionListener(this);
	    cancel.addActionListener(this);
	    buttonP.add(ok);
	    buttonP.add(cancel);

	    Container content = getContentPane();
	    content.add(textP, BorderLayout.CENTER);
	    content.add(buttonP, BorderLayout.SOUTH);
	}

	
	public void actionPerformed(ActionEvent event) {
	    if (event.getSource() == devices) 
		properties.requestFocus();
	    else if (event.getSource() == properties) 
		ok.doClick();
	    else if (event.getSource() == cancel) 
		frame.setVisible(false);
	    else if (event.getSource() == ok) {
		String s1 = devices.getText();
		String s2 = properties.getText();
		Vector v1 = new Vector();
		Vector v2 = new Vector();
		StringTokenizer st1 = new StringTokenizer(s1, ", \t\n\r\f");
		StringTokenizer st2 = new StringTokenizer(s2, ", \t\n\r\f");
		while (st1.hasMoreTokens()) 
		    v1.addElement(st1.nextToken());
		while(st2.hasMoreTokens()) 
		    v2.addElement(st2.nextToken());

		if (v1.size() == 0) {
		    optionPanel("Empty Devices", "No Input Error", 1);
		    devices.requestFocus();
		}
		else if (v2.size() == 0) {
		    optionPanel("Empty Properties", "No Input Error", 1);
		    properties.requestFocus();
		}
		else {
		    dName = new String[v1.size()];
		    pName = new String[v2.size()];
		    for(int i=0; i<v1.size(); i++)
			dName[i] = (String)v1.elementAt(i);
		    for(int i=0; i<v2.size(); i++)
			pName[i] = (String)v2.elementAt(i);
		    
		    table = new CTable(dName, pName);
		    table.addPropertyChangeListener(CTableUtility.this);
		    scroll = new JScrollPane(table);
		    scroll.setViewportView(table);
		    contentPane.add(scroll, BorderLayout.CENTER);	
		    try {  sys.flush(); }
		    catch(Exception e){ e.printStackTrace(); }
		    validate();
		    frame.setVisible(false);
		    CTableUtility.this.pack();
		}
	    }
	    
	}
    } // end newInputFrame inner class
    

    /**
     * main function
     * @param args      String array that contains the command line arguments 
     *
     * This String array can have 0 to 3 string.
     * args = null 
     * create a empty table and connect to localhost at default port.
     * args = String fileName 
     * create a table using infomation in configuration file fileName 
     * and connect to localhost at default port.
     * args = String hostName, String portNum
     * create a empty table and connect to host hostName at port portNum.
     * args = String hostName, String portNum, String fileName
     * create a table using infomation in configuration file fileName
     * and connect to host hostName at port portNum.
     */
    static public void main (String args[]){
	if (args.length == 1) {
	    (new CTableUtility(args[0])).show();
	}
	else if (args.length == 2) {
	    (new CTableUtility(args[0], args[1])).show();
	}
	else if (args.length == 3) {
	    (new CTableUtility(args[0],args[1],args[2])).show();
	}
	else {
	    (new CTableUtility(null)).show();
	}
	//--- leave one non-daemon thread (this thread) running -----
	try { Thread.currentThread().suspend(); } 
	catch (SecurityException e) {}
    }

}  // end of CTableUtility














